@@ -49,6 +49,7 @@ class MultiLayerNet:
         self.layers = OrderedDict()
         self.lastLayer = Layer.IdentityWithLoss()
         self.hiddenSizeList = []
+        self.trainable_param_keys = []
         self.prevDenseLayer = None
         self.prevConvLayer = None
 
@@ -81,6 +82,8 @@ class MultiLayerNet:
                 #                 input_size = self.prevDenseLayer.hidden_size
                 #                 print(self.hiddenSizeList)
                 input_size = int(np.prod(np.asarray(self.hiddenSizeList[-1])))
+            elif input_size is not None and len(self.hiddenSizeList) == 0:
+                input_size = int(np.prod(np.asarray(input_size)))
             else:
                 self.hiddenSizeList.append(input_size)
 
@@ -102,10 +105,19 @@ class MultiLayerNet:
             #             print(input_size)
             #             print("input", self.hiddenSizeList)
             #             print(input_size, hidden_size)
+            # print("SDFSFD", input_size, hidden_size)
+
+            # if type(input_size) == tuple:
+
             self.params[f"W{layer_len}"] = weight_init_std * np.random.randn(input_size, hidden_size)
             self.params[f"b{layer_len}"] = np.zeros(hidden_size)
 
-            self.layers[f"Affine{layer_len}"] = Layer.Affine(self.params[f"W{layer_len}"], self.params[f"b{layer_len}"])
+            self.layers[f"Affine{layer_len}"] = Layer.Affine(self.params[f"W{layer_len}"], self.params[f"b{layer_len}"],
+                                                             trainable=layer.trainable)
+
+            if layer.trainable is True:
+                self.trainable_param_keys.append(f"W{layer_len}")
+                self.trainable_param_keys.append(f"b{layer_len}")
 
             self.prevDenseLayer = layer
 
@@ -153,9 +165,9 @@ class MultiLayerNet:
 
             conv_output_size = (
                 layer.filter_num, int((input_size[1] - layer.filter_size[0] + 2 * layer.pad) / layer.stride + 1),
-            int((input_size[2] - layer.filter_size[1] + 2 * layer.pad) / layer.stride + 1))
+                int((input_size[2] - layer.filter_size[1] + 2 * layer.pad) / layer.stride + 1))
             #
-            print(conv_output_size, "!!!!!!")
+            # print(conv_output_size, "!!!!!!")
 
             self.hiddenSizeList.append(conv_output_size)
 
@@ -172,9 +184,18 @@ class MultiLayerNet:
 
             self.params[f"b{layer_len}"] = np.zeros(layer.filter_num)
             debug_print(input_size)
+
+
+
+
             self.layers[f"Convolution{layer_len}"] = Layer.Convolution(self.params[f"W{layer_len}"],
                                                                        self.params[f"b{layer_len}"], layer.stride,
-                                                                       layer.pad)
+                                                                       layer.pad, trainable=layer.get_trainable())
+
+            if layer.trainable is True:
+                self.trainable_param_keys.append(f"W{layer_len}")
+                self.trainable_param_keys.append(f"b{layer_len}")
+
             debug_print(input_size)
             self.prevConvLayer = layer
 
@@ -236,9 +257,9 @@ class MultiLayerNet:
             n = int(np.prod(np.asarray(self.hiddenSizeList[-1])))
             # print(n)
             layer.__init__(gamma=np.ones(n),
-                beta=np.zeros(n),
-                running_mean=layer.running_mean,
-                running_var=layer.running_var)
+                           beta=np.zeros(n),
+                           running_mean=layer.running_mean,
+                           running_var=layer.running_var)
             self.layers[f"BatchNormal{layer_len}"] = layer
             # self.layers[f"BatchNormal{layer_len}"] = Layer.BatchNormalization(
             #     gamma=np.ones(n),
@@ -247,6 +268,16 @@ class MultiLayerNet:
             #     running_var=layer.running_var
             # )
         # print(layer, self.hiddenSizeList)
+        elif isinstance(layer, Layer.FreezePrev):
+            for key, layer in self.layers.items():
+                if isinstance(layer, Layer.BatchNormalization) or isinstance(layer, Layer.Affine) or isinstance(layer,
+                                                                                                                Layer.Convolution):
+                    self.layers[key].trainable = False
+                    print(self.layers[key].trainable)
+
+            self.trainable_param_keys = []
+            return
+
         if is_direct:
             self.added_layer_list.append(layer)
 
@@ -257,7 +288,7 @@ class MultiLayerNet:
 
                 x = layer.forward(x, train_flg)
             else:
-
+                # print(layer)
                 x = layer.forward(x)
 
         return x
@@ -282,8 +313,8 @@ class MultiLayerNet:
 
         for i in range(math.ceil(data_num / batch_size)):
             # print("batch", i)
-            xx = x[batch_size * i: min(batch_size * (i+1), data_num)]
-            tt = t[batch_size * i: min(batch_size * (i+1), data_num)]
+            xx = x[batch_size * i: min(batch_size * (i + 1), data_num)]
+            tt = t[batch_size * i: min(batch_size * (i + 1), data_num)]
 
             yy = self.predict(xx, train_flg=False)
             yy = np.argmax(yy, axis=1)
@@ -291,12 +322,10 @@ class MultiLayerNet:
 
             correct_cnt += np.sum(yy == tt)
 
-
-
         accuracy = correct_cnt / float(x.shape[0])
         return accuracy
 
-    def gradient(self, x, t):
+    def gradient(self, x, t, train_flg = True):
         # forward
         self.loss(x, t, train_flg=True)
 
@@ -307,7 +336,8 @@ class MultiLayerNet:
         layers = list(self.layers.values())
         layers.reverse()
         for layer in layers:
-            dout = layer.backward(dout)
+            if layer.trainable is True or train_flg is False:
+                dout = layer.backward(dout)
 
         # 결과 저장
         grads = {}
@@ -341,6 +371,8 @@ class MultiLayerNet:
             else:
                 raise "err"
 
+        optimizer.__init__(trainable_param_keys=self.trainable_param_keys)
+
         iters_num = kwargs.get("iters_num", None)
         if iters_num is None:
             return ValueError("iters_num must be set")
@@ -389,13 +421,14 @@ class MultiLayerNet:
         test_acc_list = []
 
         iter_per_epoch = max(train_size / batch_size, 1)
+        print("iter_per_epoch :", iter_per_epoch)
 
         print(f"repeat {int(iters_num // (max(train_size / batch_size, 1)))} epoch")
 
         if output_type == "regression":
             print("epoch | loss")
         else:
-            print("epoch | train_acc | test_acc")
+            print("epoch | train_acc | test_acc | loss")
 
         cnt = 1
 
@@ -414,7 +447,7 @@ class MultiLayerNet:
             x_batch = np.asarray(x_train[batch_mask])
             t_batch = np.asarray(t_train[batch_mask])
 
-            grad = self.gradient(x_batch, t_batch)
+            grad = self.gradient(x_batch, t_batch, train_flg = False)
 
             # 갱신
 
@@ -440,6 +473,7 @@ class MultiLayerNet:
                     else:
                         # log.info(f"epoch {cnt}:")
                         print(f"epoch {cnt}:", format(float(train_acc), ".4f"), " | ", format(float(test_acc), ".4f"),
+                              " | ", format(float(loss), ".4f"),
                               flush=True)
 
                 cnt += 1
@@ -495,8 +529,6 @@ class MultiLayerNet:
                 other_params[f"BN_m{layer_idx}"] = layer.running_mean
                 other_params[f"BN_v{layer_idx}"] = layer.running_var
 
-
-
         save_data["LayerInfo"] = layer_info
         # for layer_idx, layer in enumerate(self.layers.values()):
         #     if isinstance(layer, Layer.BatchNormalization):
@@ -513,6 +545,7 @@ class MultiLayerNet:
         np.savez_compressed(path, **save_data)
         print(f"Weight was saved at {path}")
 
+
     def load_model(self, path):
         import numpy as np2
         print(f"Loading Model... {path}")
@@ -522,6 +555,12 @@ class MultiLayerNet:
         layer_info = load_data.get("LayerInfo")
         for layer_encoded in layer_info:
             layer = jsp.decode(layer_encoded)
+            if hasattr(layer, "trainable"):
+                pass
+            else:
+                layer.trainable = True
+
+
             # print(layer_encoded)
 
             self.add_layer(layer)
@@ -534,7 +573,13 @@ class MultiLayerNet:
 
             layer_idx = int(re.findall(r'[0-9]+', param_name)[0])
             param_kind = re.sub(r'[0-9]+', '', param_name)
+
+
+
             target_layer = list(self.layers.items())[layer_idx][1]
+
+            print(layer_idx, param_kind, target_layer)
+
             if param_kind == "W":
                 target_layer.W = param_value
             elif param_kind == "b":
@@ -556,8 +601,6 @@ class MultiLayerNet:
             elif param_kind == "BN_v":
                 target_layer.running_var = param_value
 
-
-
 # %%
 
 
